// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

#include <string>
#include <sstream>

namespace onnxruntime {
namespace webgpu {

std::string UtilFunctions(std::string stride_string);

}  // namespace webgpu
}  // namespace onnxruntime
