// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License

#pragma once

#include <hip/hip_runtime.h>
#include <iso646.h>
#include <migraphx/migraphx.hpp>
#include <migraphx/version.h>
