/*
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 *
 * Generated by generate_moe_gemm_tma_ws_sm90_fp4.py. Do not edit manually.
 */

#ifdef COMPILE_HOPPER_TMA_GROUPED_GEMMS
#if defined(ENABLE_FP4) && defined(USE_FP4_QMOE)
#ifndef ORT_QUICK_BUILD
#ifdef ENABLE_BF16
#include "contrib_ops/cuda/llm/moe_gemm/launchers/moe_gemm_tma_ws_sm90_fp4_instantiation.cuh"

namespace onnxruntime::llm::kernels::cutlass_kernels {

ORT_MOE_GEMM_TMA_WS_SM90_FP4_INST_PP_FINALIZE(__nv_bfloat16, 64, 16, 128, 1, 1, 1);

}  // namespace onnxruntime::llm::kernels::cutlass_kernels

#endif  // ENABLE_BF16
#endif  // !ORT_QUICK_BUILD
#endif  // ENABLE_FP4 && USE_FP4_QMOE
#endif  // COMPILE_HOPPER_TMA_GROUPED_GEMMS
