import 'onnxruntime-web/ort-wasm-simd-threaded.wasm?url&no-inline';
import 'onnxruntime-web/ort-wasm-simd-threaded.mjs?url&no-inline';

import 'onnxruntime-web/ort-wasm-simd-threaded.jsep.wasm?url&no-inline';
import 'onnxruntime-web/ort-wasm-simd-threaded.jsep.mjs?url&no-inline';

import 'onnxruntime-web/ort-wasm-simd-threaded.jspi.wasm?url&no-inline';
import 'onnxruntime-web/ort-wasm-simd-threaded.jspi.mjs?url&no-inline';

import 'onnxruntime-web/ort-wasm-simd-threaded.asyncify.wasm?url&no-inline';
import 'onnxruntime-web/ort-wasm-simd-threaded.asyncify.mjs?url&no-inline';

import './main.js';
