// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

package ai.onnxruntime.reactnative;

import ai.onnxruntime.extensions.OrtxPackage;

class OnnxruntimeExtensions {
  static public String getLibraryPath() {
    return OrtxPackage.getLibraryPath();
  }
}
